#clear all variables in the memory
rm(list = ls())


############# PART 1 - prediction intervals ####################
library(faraway)
data(fat,package="faraway")

#remove outliers
#fat$abdom[order(fat$abdomen)]
#fat$brozek[order(fat$brozek)]

#39 - abdomen circumference 148cm, 164kg
#41 - abdomen circumference 126cm
#182 - 0% body fat = dead person
#172 - 1.9% body fat = also dead person
fat <- fat[-c(39,41,182,172),]

#estimate linear model with these regressors
lmod <- lm(brozek ~ age + weight + height + abdom, data=fat)
x <- model.matrix(lmod)

#look at a median person in the sample

############(TO BE ADDED)###############
#save median person (persons are rows in matrix x) to x0
x0 <- 
########################################

#or look at a particular quantile
(x0 <- apply(x,2,quantile, probs= seq(0,1,0.1))[2,])

#look at persons that are "close" to x0

#create a matrix where first row is x0 and then other agents
dataX0 <- do.call(rbind, list(x0[-1],fat[,c(4,5,6,11)]))
#find out the distance
distData <- dist(dataX0, method = "euclidean", diag = FALSE, upper = FALSE, p = 2)
#it gives us n(n-1)/2 combinations, we're interested in the distance of x0 from other observations
#first n members is the distance from x0, first order it from closest to the most distant 
orderVec <- order(distData[1:dim(fat)[1]])
#save the first k observations that are "close" to x0
k <- 50
############(TO BE ADDED)###############
#save the k observations that are closest to fatClose
#which will be a data.frame
fatClose <- 
########################################

#we are about to plot the relationship between abdom and brozek
#store the quantiles of abdomen
abdomQuant <- quantile(fat$abdom,seq(0,1,0.1))
abdomMatrix <- rep(1,11) %*% t(x0)
abdomMatrix[,5] <- abdomQuant
#now in abdomMatrix, there is person with median predictors, but abdom

#create a function for a prediction interval
predInt <- function(x0){
  return(predict(lmod,new = data.frame(t(x0)),interval = "prediction"))
}

############(TO BE ADDED)###############
#create a function for a prediction interval of a mean of a future value
#this function will be called conInt
conInt <- function(x0){
  return()
}
########################################

#apply these functions to the rows of abdomMatrix
#(faster than for cycle)
pM <- apply(abdomMatrix, 1, predInt)
cM <- apply(abdomMatrix, 1, conInt)

#merge these matrices so we can plot them easily
mergedM <- do.call(rbind, list(abdomQuant,pM,cM))
df <- data.frame(t(mergedM))

library(ggplot2)
#plot separately point that are "close" to x0
ggplot(df, aes(x=X1)) +
  geom_line(aes(y=X2), size = 1.5)+
  geom_line(aes(y=X6),colour = "grey", size = 1.5)+ geom_line(aes(y=X7), colour = "grey", size = 1.5)+
  geom_line(aes(y=X3), colour = "grey", linetype = "dashed", size = 1.5)+ geom_line(aes(y=X4), colour = "grey", linetype = "dashed", size = 1.5)+
  geom_point(data = fat, aes(abdom,brozek), colour="grey")+
  ############(TO BE ADDED)###############
  #add red points of size 2 as those points who are "close" to x0
  #use data = fatClose for that
  
  ########################################
  labs(title = "Prediction Intervals",
       x = "Abdom",
       y = "Fat %")



############# PART 2 - cross-validation ####################

#https://www.r-bloggers.com/cross-validation-for-predictive-analytics-using-r/


# Generate the training and test samples
seed <- 1809
set.seed(seed)

#generate artificial data according to parameter beta
gen_data <- function(n, beta, sigma_eps) {
  ############(TO BE ADDED)###############
  #generate a random vector of size n of iid N(0,sigma_eps^2)
  #call this vector eps
  eps <- 
  ########################################
  x <- sort(runif(n, 0, 100))
  X <- cbind(1, poly(x, degree = (length(beta) - 1), raw = TRUE))
  y <- as.numeric(X %*% beta + eps)
  return(data.frame(x = x, y = y))
}

# Fit the models
require(splines)

#true parameter beta is 4-dimensional
n_rep <- 100
n_df <- 30
df <- 1:n_df
beta <- c(5, -0.1, 0.004, -3e-05)
n_train <- 50
n_test <- 10000
sigma_eps <- 0.5

xy <- res <- list()
############(TO BE ADDED)###############
#use gen_data function to generate test sample with 
#size n_test and parameters beta and sigma_eps
xy_test <- 
########################################
for (i in 1:n_rep) {
  xy[[i]] <- gen_data(n_train, beta, sigma_eps)
  x <- xy[[i]][, "x"]
  y <- xy[[i]][, "y"]
  #estimate
  #ns generates nbasis of natural cubic spline
  res[[i]] <- apply(t(df), 2, function(degf) lm(y ~ ns(x, df = degf)))
}

# Plot the data
x <- xy[[1]]$x
X <- cbind(1, poly(x, degree = (length(beta) - 1), raw = TRUE))
y <- xy[[1]]$y
plot(y ~ x, col = "gray", lwd = 2)
lines(x, X %*% beta, lwd = 3, col = "black")
lines(x, fitted(res[[1]][[1]]), lwd = 3, col = "palegreen3")
lines(x, fitted(res[[1]][[4]]), lwd = 3, col = "darkorange")
############(TO BE ADDED)###############
#plot line graph of fitted values of the polynomial of degree 25
#use color = "steelblue"

########################################
legend(x = "topleft", legend = c("True function", "Linear fit (df = 1)", "Best model (df = 4)", 
                                 "Overfitted model (df = 25)"), lwd = rep(3, 4), col = c("black", "palegreen3", 
                                                                                         "darkorange", "steelblue"), text.width = 32, cex = 0.85)

# n-fold cross-validation
set.seed(seed)

n_train <- 100
xy <- gen_data(n_train, beta, sigma_eps)
x <- xy$x
y <- xy$y

fitted_models <- apply(t(df), 2, function(degf) lm(y ~ ns(x, df = degf)))
mse <- sapply(fitted_models, function(obj) deviance(obj)/nobs(obj))

n_test <- 10000
xy_test <- gen_data(n_test, beta, sigma_eps)
pred <- mapply(function(obj, degf) predict(obj, data.frame(x = xy_test$x)), 
               fitted_models, df)
te <- sapply(as.list(data.frame(pred)), function(y_hat) mean((xy_test$y - y_hat)^2))

n_folds <- 10
folds_i <- sample(rep(1:n_folds, length.out = n_train))
cv_tmp <- matrix(NA, nrow = n_folds, ncol = length(df))
for (k in 1:n_folds) {
  test_i <- which(folds_i == k)
  train_xy <- xy[-test_i, ]
  test_xy <- xy[test_i, ]
  x <- train_xy$x
  y <- train_xy$y
  fitted_models <- apply(t(df), 2, function(degf) lm(y ~ ns(x, df = degf)))
  x <- test_xy$x
  y <- test_xy$y
  pred <- mapply(function(obj, degf) predict(obj, data.frame(ns(x, df = degf))), 
                 fitted_models, df)
  cv_tmp[k, ] <- sapply(as.list(data.frame(pred)), function(y_hat) mean((y - 
                                                                           y_hat)^2))
}
cv <- colMeans(cv_tmp)


#install.packages("Hmisc")
require(Hmisc)

plot(df, mse, type = "l", lwd = 2, col = gray(0.4), ylab = "Prediction error", 
     xlab = "Flexibilty (spline's degrees of freedom [log scaled])", main = paste0(n_folds, 
                                                                                   "-fold Cross-Validation"), ylim = c(0.1, 0.8), log = "x")
lines(df, te, lwd = 2, col = "darkred", lty = 2)
cv_sd <- apply(cv_tmp, 2, sd)/sqrt(n_folds)
errbar(df, cv, cv + cv_sd, cv - cv_sd, add = TRUE, col = "steelblue2", pch = 19, 
       lwd = 0.5)
lines(df, cv, lwd = 2, col = "steelblue2")
points(df, cv, col = "steelblue2", pch = 19)
legend(x = "topright", legend = c("Training error", "Test error", "Cross-validation error"), 
       lty = c(1, 2, 1), lwd = rep(2, 3), col = c(gray(0.4), "darkred", "steelblue2"), 
       text.width = 0.4, cex = 0.85)



#leave one out cross validation
require(splines)

loocv_tmp <- matrix(NA, nrow = n_train, ncol = length(df))
for (k in 1:n_train) {
  train_xy <- xy[-k, ]
  test_xy <- xy[k, ]
  x <- train_xy$x
  y <- train_xy$y
  fitted_models <- apply(t(df), 2, function(degf) lm(y ~ ns(x, df = degf)))
  pred <- mapply(function(obj, degf) predict(obj, data.frame(x = test_xy$x)),
                 fitted_models, df)
  loocv_tmp[k, ] <- (test_xy$y - pred)^2
}
loocv <- colMeans(loocv_tmp)

plot(df, mse, type = "l", lwd = 2, col = gray(.4), ylab = "Prediction error",
     xlab = "Flexibilty (spline's degrees of freedom [log scaled])",
     main = "Leave-One-Out Cross-Validation", ylim = c(.1, .8), log = "x")
lines(df, cv, lwd = 2, col = "steelblue2", lty = 2)
lines(df, loocv, lwd = 2, col = "darkorange")
legend(x = "topright", legend = c("Training error", "10-fold CV error", "LOOCV error"),
       lty = c(1, 2, 1), lwd = rep(2, 3), col = c(gray(.4), "steelblue2", "darkorange"),
       text.width = .3, cex = .85)




